#!/usr/bin/python3

from threading import Thread
from time import sleep
from queue import Queue
from xml.etree import ElementTree
from urllib.parse import urljoin
from os.path import isfile, dirname, join
from os import remove, system
import pytest
import requests
import json


class TestStreamingAlerts(object):
    """
    A testing class used to group all the tests related to the streaming scenario.
    """

    kapacitor_url = "http://localhost:9092/kapacitor/v1/tasks"

    @pytest.mark.parametrize("rule, log", [
        ("rules.json", "/tmp/RPSLoad.log"),
        ("test_rule1.json", "/tmp/TestRule1.log"),
        ("test_rule2.json", "/tmp/TestRule2.log"),
    ])
    def test_alerts(self, rule, log, streaming_url, streaming_manifest):
        """
        This test case generates some streaming requests to the server to ensure an alert is triggered and then tests the log file for this alert. Different logs can be tested by
        appending to the list of parameters in the pytest decorator.

        Format for pytest parameters under test:
        ([filename], [log])
        where [filename] is the name of the json file for the rule under test (must be in the same folder as this test is)
              [log] is the absolute path of the log file that must be created due to an alert

        :param rule: the name of the rule json file
        :param log: the path of the log file that is under test
        :param streaming_url: the fixture providing the streaming url for this test case
        :param streaming_manifest: the fixture providing the root of the XML streaming manifest
        """

        kapacitor_setter = self.kapacitor_setting(rule, log)
        next(kapacitor_setter)  # Setup the test rule

        print("Testing alert creation for rule: {0}".format(rule))

        segments = streaming_manifest.findall(".//{urn:mpeg:DASH:schema:MPD:2011}SegmentURL")

        threads_num = 30
        threads_queue = Queue(maxsize=threads_num)  # a synchronized queue is used to track if all the threads has finished execution
        threads = [StreamingThread(streaming_url, segments, threads_queue) for _ in range(threads_num)]
        for t in threads:
            t.start()

        alert_created = False
        counter = 0
        time_delay = 2.5
        while True:
            # loop while threads are execution and do a check every 2.5 seconds to check if either alert log has been created or threads have finished execution
            sleep(time_delay)
            if isfile(log):
                for t in threads:  # kill all running threads in case log file is created beforehand
                    t.stop()
                alert_created = True

            if threads_queue.full():
                break

            counter += time_delay  # the counter tracks the time taken, for the rules under test usually a 20 seconds time frame is enough
            if counter >= 8*time_delay:
                for t in threads:  # kill all running threads in case of test failure
                    t.stop()
                break

        assert alert_created, "Alerts test failed: no log file is created indicating a triggered alert for rule {0}.".format(rule)

        print("Successfully passed alert creation test for rule: {0}.".format(rule))

        next(kapacitor_setter)  # Teardown the test rule

    def kapacitor_setting(self, rule, log):
        """
        A generator function used to provide setUp/tearDown actions for a particular kapacitor rule.
        On setUp rule is initialized, on tearDown rule is deleted. Interleaving is achieved using the generator pattern.

        :param rule: the name of the json file for the rule under test
        :param log: the absolute path of the log file that's being tested
        """

        # check if the log file is already created due to a previous test
        try:
            if isfile(log):
                remove(log)  # delete log file if existing from previous tests
        except PermissionError:
            system("sudo rm {0}".format(log))  # handles the case for running on linux where permission will be required to delete the old log file

        # Initialization of the kapacitor rule - Test setUp (UnitTest style)
        with open(join(dirname(__file__), rule), "r") as rule_file:
            data = "".join(line.strip() for line in rule_file.readlines())

        rule_data = json.loads(data)
        requests.delete(url=urljoin(self.kapacitor_url + "/", rule_data.get("id")))  # delete in case of a task with the same ID already set in the kapacitor
        r = requests.post(url=self.kapacitor_url, data=data, headers={"Content-Type": "application/json"})
        assert r.status_code == 200, "Couldn't create alert rule {0}".format(rule)
        print("\nSuccessfully created test rule {0}".format(rule))

        yield

        # Deleting the kapacitor rule used for testing - Test tearDown (UnitTest style)
        requests.delete(url=urljoin(self.kapacitor_url + "/", rule_data.get("id")))

        # check if the log file is created and clean it up
        try:
            if isfile(log):
                remove(log)  # delete log file if existing from previous tests
        except PermissionError:
            system("sudo rm {0}".format(log))  # handles the case for running on linux where permission will be required to delete the old log file
        yield

    @staticmethod
    @pytest.fixture(scope="class", params=[{"server": "http://192.168.50.11", "video": "/test_video/stream.mpd"}])
    def streaming_url(request):
        """
        A fixture with class scope - used only in the scope of the testing class.

        :param request: the parameters for this fixture - server url and video relative url
        :return: the combined URL for the video used for streaming
        """

        return urljoin(request.param["server"], request.param["video"])

    @staticmethod
    @pytest.fixture(scope="class")
    def streaming_manifest(streaming_url):
        """
        A fixture to download the manifest file for the streamed video and parse the downloaded XML content

        :param streaming_url: the fixture which provides the streaming url
        :return: an XML root node object
        """

        manifest_xml = requests.get(streaming_url).text
        root = ElementTree.fromstring(manifest_xml)
        return root


class StreamingThread(Thread):

    def __init__(self, url, segments, queue):
        """
        Subclassing the Thread class to create a custom streaming thread.

        :param url: the streaming url
        :param segments: the list of SegmentURL XML nodes
        :param queue: an auxiliary parameter used to indicate when this thread has finished execution
        """

        super(StreamingThread, self).__init__()
        self.running = False
        self.url = url
        self.segments = segments
        self.queue = queue
        self._test_finished = False  # a flag to indicate whether the thread should stop running

    def stop(self):
        """
        Kill this thread and suspend its execution.
        """

        self._test_finished = True

    def run(self):
        """
        A function, which simulates an actual streaming by downloading different audio/video segments from the server using a request session,
        which leaves the connection open until executing.
        """

        size = len(self.segments)
        size = size if size % 2 == 0 else size - 1

        s = requests.session()

        for i in range(0, int(size / 2), 1):
            segment_audio = self.segments[0]
            segment_video = self.segments[int(size / 2) + i]
            segment_audio_url = segment_audio.attrib.get('media')
            segment_video_url = segment_video.attrib.get('media')

            s.get(urljoin(self.url, segment_audio_url))
            s.get(urljoin(self.url, segment_video_url))

            # check if thread is killed in case the test has already succeeded
            if self._test_finished:
                break

            # a small time out to mimic the behaviour of a real streaming
            sleep(2.5)

        self.queue.put(True)
